
%% Combining images of different exposure
%% Richard Mann, Jan. 2001

%% Reference:
%% Debevec, P.E. and Malik, J.  Siggraph 1997.
%%
%% See original solution of problem:
%% Mann, S. and Picard R.W.  Proceedings of IS&T 46th annual conference, 1995.

%% Paste the following into Matlab window.

%% g(1:256) is ln.f^-1 log inverse imaging function defined on 0..255
%% Z(n,p) is measured image brightness, n=1..N, p=1..P
%% lE is log Brightness for each pixel 1..N
%% lT is log Exposure time for each frame 1..P
%% w(1:256) is weighting function for pixels 0..255

%% Read in P images of the same scene, with different exposure
P=3; % number of frames to use
rows=300; cols=448; % image sizes
images=cell(P,1); % cell array of images
for i=1:P
  fn=sprintf('wyck-%1d.tif', i);
  images{i}=imread(fn,'tiff');
end
for i=1:P,
  figure;
  colormap gray; imagesc(images{i},[0 255]);
  str=sprintf('Image %d',i);
  title(str);
end

%% Exposure times.  Reference time is 1 unit.
%T=[2 1 0.5];
T=[2 1 0.5/sqrt(2)];   %%%%% New value for dark image %%%%%
lT=log(T);

%% Select N random points from each of P images
N=200; % number of pixels to use
Z=zeros(N,P);
indices=floor(rand(N,1)*rows*cols+1);
for i=1:P
  Im=reshape(images{i},rows*cols,1); % make column vector
  Z(:,i)=double(Im(indices));
end

%% Triangular weighting function accentuates pixels in center of domain
n=256; % number of grey levels
w=[1:n/2,n/2:-1:1]/(n/2); w=w'; % weighting function
figure; plot(1:n,w);

%% Show scattergram of two image's pixels.
%% This is for interest only.  Note how each image saturates.
figure;
i=3; j=1;
scatter(Z(1:N,i)+randn(N,1),Z(1:N,j)+randn(N,1),'.');
axis('square');
xlabel('Im_i (dark)');
ylabel('Im_j (light)');
title('Scattergram (pixel values): Im_i vs Im_j');

%% Data fitting equations
A=zeros(N*P+n+1,n+N);
b=zeros(N*P+n+1,1);
k=1;
for i=1:N
  for j=1:P
    wij=w(Z(i,j)+1);  % pixels 0..255
    A(k,Z(i,j)+1)=wij; A(k,n+i)=-wij;
    b(k)=wij*lT(j);
    k=k+1;
  end
end

%% Fix the curve by setting middle value to zero
A(k,129)=1;
k=k+1;

%% Smoothness terms
lambda=100;
for i=1:n-2
  A(k,i)=lambda*w(i+1);  A(k,i+1)=-2*lambda*w(i+1);  A(k,i+2)=lambda*w(i+1);
  k=k+1;
end

%% Solve using SVD

x=A\b;
g=x(1:n);
lE=x(n+1:size(x,1));

figure;
plot(g,1:n);
title('Response function');
xlabel('log exposure');
ylabel('pixel value');
hold on;
plot(lE,Z(:,floor(P/2)+1),'o'); % data

%% Combine images
%% Note: code is "vectorized" below for fast operation
logEsum=zeros(rows*cols,1);
Wsum=zeros(rows*cols,1);
for i=1:P
  Im=reshape(images{i},rows*cols,1); % make column vector
  logEsum=logEsum+w(Im+1).*(g(Im+1)-lT(i));
  Wsum=Wsum+w(Im+1);
end
logE=logEsum./Wsum;
logE=reshape(logE,rows,cols);

%% Display with logistic/sigmoid tonemap
alpha=1.0;  % possible values 0.5, 1.0, 2.0
Edisp=ones(rows,cols)./(ones(rows,cols)+exp(-alpha*logE));
figure; colormap(gray); imagesc(Edisp,[0,1]);
title(sprintf('Composite image (logistic tonemap, alpha=%g)', alpha));

%% Display tonemap (compare to response function above)
SlE=ones(N,1)./(ones(N,1)+exp(-alpha*lE));
figure; plot(lE, SlE,'o');
title(sprintf('Tonemap (logistic function, alpha=%g)', alpha));
xlabel('log exposure'); ylabel('pixel value [0..1]');
axis([-4 4 0 1]);


